/*
 * webserver.cpp
 *
 *  Created on: Dec 29, 2019
 *      Author: rsn
 */
#define GLOBAL

#include "includes.h"
#include "defines.h"
#include "projStruct.h"
#include "globals.h"

extern void pprintf(const char * format, ...);
extern int sendMsg(int cualSock,uint8_t *lmessage, uint16_t son,uint8_t *donde,uint16_t maxx);

static esp_err_t challenge_get_handler(httpd_req_t *req);
static esp_err_t conn_login(httpd_req_t *req);
//static esp_err_t conn_firmware(httpd_req_t *req);
//static esp_err_t conn_tariff(httpd_req_t *req);
static esp_err_t conn_base(httpd_req_t *req);
static esp_err_t conn_menu(httpd_req_t *req);
static esp_err_t conn_setup(httpd_req_t *req);
static esp_err_t conn_config_handler(httpd_req_t *req);
static esp_err_t conn_status(httpd_req_t *req);
static esp_err_t conn_stats(httpd_req_t *req);
static esp_err_t sendUtils(httpd_req_t *req);

void shaMake(char * payload,uint8_t payloadLen, unsigned char *lshaResult);

extern const unsigned char answer_start[] 		asm("_binary_challenge_html_start");
extern const unsigned char setup_start[] 		asm("_binary_connSetup_html_start");
extern const unsigned char login_start[] 		asm("_binary_login_html_start");
extern const unsigned char final_start[] 		asm("_binary_ok_html_start");
extern const unsigned char menu_start[] 		asm("_binary_menu_html_start");
extern const unsigned char status_start[] 		asm("_binary_connStatus_html_start");
extern const unsigned char stats_start[] 		asm("_binary_connStats_html_start");
extern const unsigned char cssstyles_start[] 	asm("_binary_styles_css_start");
extern const unsigned char cssstyles_end[] 		asm("_binary_styles_css_end");

extern const unsigned char ok_start[] 			asm("_binary_ok_png_start");
extern const unsigned char ok_end[] 			asm("_binary_ok_png_end");
extern const unsigned char nak_start[] 			asm("_binary_nak_png_start");
extern const unsigned char nak_end[] 			asm("_binary_nak_png_end");
extern const unsigned char logo_start[] 		asm("_binary_meter_png_start");
extern const unsigned char logo_end[] 			asm("_binary_meter_png_end");
extern const unsigned char fav_start[] 			asm("_binary_favicon_ico_start");
extern const unsigned char fav_end[] 			asm("_binary_favicon_ico_end");

int ok_bytes=ok_end-ok_start;
int nak_bytes=nak_end-nak_start;
int logo_bytes=logo_end-logo_start;
int csslen=cssstyles_end-cssstyles_start;
int fav_bytes=fav_end-fav_start;

extern void delay(uint32_t a);
extern void write_to_flash();
extern uint32_t millis();

typedef struct httpd_uri ur;
#define MAXURLS 13
ur urls[MAXURLS];

void init_urls()
{
    urls[ 0].uri       = "/styles.css";			urls[ 0].handler   = sendUtils;
    urls[ 1].uri       = "/cmconfig";			urls[ 1].handler   = conn_config_handler;
    urls[ 2].uri       = "/cmstatus";			urls[ 2].handler   = conn_status;
    urls[ 3].uri       = "/cmstats";			urls[ 3].handler   = conn_stats;
    urls[ 4].uri       = "/cmsetup";			urls[ 4].handler   = conn_setup;
    urls[ 5].uri       = "/cmchallenge";		urls[ 5].handler   = challenge_get_handler;
    urls[ 6].uri       = "/cmvalidate";			urls[ 6].handler   = conn_login;
    urls[ 7].uri       = "/cmmenu";				urls[ 7].handler   = conn_menu;
    urls[ 8].uri       = "/favicon.ico";		urls[ 8].handler   = sendUtils;
    urls[ 9].uri       = "/ok.png";				urls[ 9].handler   = sendUtils;
    urls[10].uri       = "/nak.png";			urls[10].handler   = sendUtils;
    urls[11].uri       = "/meter.png";			urls[11].handler   = sendUtils;
    urls[12].uri       = "/";					urls[12].handler   = conn_base;

}


int sendHtml(httpd_req_t *req,const char * format, ...)
{
	va_list args;
	char *ltemp;

	ltemp=(char*)malloc(5000);
	if(!ltemp)
		return ESP_FAIL;
	else
		bzero(ltemp,5000);

	va_start (args, format);
	vsprintf (ltemp,format, args);
	va_end (args);

	httpd_resp_send(req,ltemp, strlen(ltemp));
	FREEANDNULL(ltemp);
	return ESP_OK;
}


void monitorCallback( TimerHandle_t xTimer )
{
	webState=wNONE;
//cannot send a message directly so general error type msg

}

esp_err_t sendUtils(httpd_req_t *req)
{
	unsigned char * theP=NULL;
	int theLen=0;


	httpd_resp_set_hdr(req,"Cache-Control","private, max-age=86400");
	httpd_resp_set_type(req,"image/png");
	if(strcmp(req->uri,"/favicon.ico")==0)
	{
		theP=(unsigned char*)fav_start;
		theLen=fav_bytes;
	}
	else
		if(strcmp(req->uri,"/nak.png")==0)
		{
			theP=(unsigned char*)nak_start;
			theLen=nak_bytes;
		}
		else
			if(strcmp(req->uri,"/ok.png")==0)
			{
				theP=(unsigned char*)ok_start;
				theLen=ok_bytes;
			}
			else
				if(strcmp(req->uri,"/meter.png")==0)
				{
					theP=(unsigned char*)logo_start;
					theLen=logo_bytes;
				}
				else
					if(strcmp(req->uri,"/styles.css")==0)
					{
						theP=(unsigned char*)cssstyles_start;
						theLen=csslen;
						httpd_resp_set_type(req,"text/css");
					}
	httpd_resp_send(req,(char*)theP,theLen);
	return ESP_OK;
}



int getSock()
{
	ip_addr_t 						remote;
	u8 								aca[4];
	tcpip_adapter_ip_info_t 		ip_info;
	int 							addr_family,lsock;
	int								ip_protocol,err;

	ESP_ERROR_CHECK(tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_STA, &ip_info));


	ESP_ERROR_CHECK(tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_STA, &ip_info));
	memcpy(&aca,&ip_info.gw,4);
	IP_ADDR4( &remote, aca[0],aca[1],aca[2],aca[3]);

	struct sockaddr_in dest_addr;
	memcpy(&dest_addr.sin_addr.s_addr,&ip_info.gw,4);	//send to our gateway, which is our server
	dest_addr.sin_family = AF_INET;
	dest_addr.sin_port = htons(BDGHOSTPORT);
	addr_family = AF_INET;
	ip_protocol = IPPROTO_IP;

	lsock =  socket(addr_family, SOCK_STREAM, ip_protocol);
	if (lsock < 0) {
		pprintf( "Unable to create socket: errno %d\n", errno);
		return ESP_FAIL;

	}

	err = connect(lsock, (struct sockaddr *)&dest_addr, sizeof(dest_addr));
	if (err != 0)
	{
			pprintf( "%sSocket %d unable to connect: errno %d\n",WIFIDT,gsock, errno);
			return ESP_FAIL;
	}
	return lsock;
}

cJSON *makeJson()
{

	cJSON *ar = cJSON_CreateArray();
	if(!ar)
	{
		pprintf("Error creating rsvp arr\n");
		return NULL;
	}

	cJSON *root=cJSON_CreateObject();
	if(!root)
	{
		pprintf("Error creating rsvp cjson arr\n");
		return NULL;
	}

	for (int a=0;a<MAXDEVS;a++)
	{
		if(theConf.configured[a]==1)
		{
				cJSON_AddStringToObject(root,"mid",				theConf.medidor_id[a]);
				cJSON_AddNumberToObject(root,"KwH",				theConf.bornKwh[a]);
				cJSON_AddNumberToObject(root,"BPK",				theConf.beatsPerKw[a]);
		}
	}
	cJSON_AddStringToObject			(root,"cmd","cmd_rsvp");
	cJSON_AddItemToArray(ar, root);

	return ar;
}

int reserveSlot()		//send ConnMgr a request for Whitelisting this MAC
{
#define MSGSIZE	500
	int q=0,misock;

	if(strlen(connserver)==0)
	{
		printf("No connmgr %s\n",connserver);
		return ESP_FAIL;
	}

	if(theConf.active)
		misock=gsock;
	else
		misock=getSock();
	if(misock<0)
	{
		printf("No socket %d\n",errno);
		return ESP_FAIL;
	}

	cJSON *root=cJSON_CreateObject();
	if(root==NULL)
	{
		pprintf("cannot create root rsvp\n");
		return ESP_FAIL;
	}

	cJSON *inner=makeJson();
	if(!inner)
	{
		cJSON_Delete(root);
		return ESP_FAIL;
	}

	cJSON_AddStringToObject			(root,"Controller",theConf.meterName);
	cJSON_AddStringToObject			(root,"cmd","cmd_rsvp");
	cJSON_AddNumberToObject			(root,"macn",(double)theMacNum);
	cJSON_AddStringToObject			(root,"user",username);
	cJSON_AddStringToObject			(root,"passw",password);
	cJSON_AddItemToObject			(root,"Batch",inner);

	char *lmessage=cJSON_Print(root);
	if(!lmessage)
	{
		cJSON_Delete(root);
		return ESP_FAIL;
	}

	cJSON_Delete(root);

	uint8_t *ansmem=(uint8_t*)malloc(MSGSIZE);
	if(!ansmem)
	{
		pprintf("No heap2\n");
		FREEANDNULL(lmessage);
		return ESP_FAIL;
	}
	bzero(ansmem,MSGSIZE);

	if(global_crypt)
		theConf.crypt=1;	// encryption now
	else
		theConf.crypt=0;	//fresh setup no additions

	q=sendMsg(misock,(uint8_t*)lmessage,strlen(lmessage),(uint8_t*)ansmem,MSGSIZE);
	if(q<0)
	{
		if(theConf.traceflag & (1<<WEBD))
			printf("%sFail to send %d scok %d\n",WEBDT,q,misock);
		FREEANDNULL(ansmem)
		FREEANDNULL(lmessage)
		return ESP_FAIL;
	}
	if(theConf.traceflag & (1<<WEBD))
		printf("%sRsvp successfull [%s]\n",WEBDT,ansmem);
	FREEANDNULL(ansmem)
	FREEANDNULL(lmessage)
	shutdown(misock,0);
	close(misock);

	return ESP_OK;
}

static int scan(char * who)
{
	 	uint16_t number = 10;
	    wifi_ap_record_t ap_info[10];
	    uint16_t ap_count = 0;
	    int lowest=-1000;

	    bzero(ap_info, sizeof(ap_info));

	    wifi_scan_config_t scan_config;
		scan_config.ssid = 0;
		scan_config.bssid = 0;
		scan_config.channel = 0;
		scan_config.show_hidden = true;
		scan_config.scan_type= WIFI_SCAN_TYPE_ACTIVE;
		scan_config.scan_time.active.max=0;	//120ms per channel. The fastest
		scan_config.scan_time.active.min=0;

	   	ESP_ERROR_CHECK(esp_wifi_scan_start(&scan_config, true));
	    ESP_ERROR_CHECK(esp_wifi_scan_get_ap_records(&number, ap_info));
	    ESP_ERROR_CHECK(esp_wifi_scan_get_ap_num(&ap_count));

	    if(theConf.traceflag & (1<<WEBD))
	    	printf("%sTotal APs scanned = %u%s\n",RED, ap_count,RESETC);

	    if(ap_count>0)
	    {
	    for (int i = 0; (i < 10) && (i < ap_count); i++)
	    {
	    	if(ap_info[i].rssi>lowest)
	    	{
				if(strstr((char *)ap_info[i].ssid,"CmgrIoT")!=NULL)
				{
					memcpy(who,&ap_info[i].ssid,strlen((char*)ap_info[i].ssid));
					if(theConf.traceflag & (1<<WEBD))
					{
						pprintf("%sSSID%s \t\t%s\n",GREEN,RESETC, ap_info[i].ssid);
						pprintf("%sRSSI%s\t\t%d\n", CYAN,RESETC,ap_info[i].rssi);
					}
				}
	    	}
	    }
	    	return ESP_OK;
	    }
	    else
	    	return ESP_FAIL;
}


void resetlater(void *parg)
{
	delay(3000);
	esp_restart();
}

static esp_err_t challenge_get_handler(httpd_req_t *req)
{
#define SHAC	16
#define SHAL	(SHAC/2)-1
	bool		rebf=false;		//reboot flag
    char*  		buf;
    size_t 		buf_len;
	char 		param[100];

//	   if(webState!=wCHALL)
//		{
//			sprintf(tempb,(char*)final_start,"nak","Invalid Control Sequence or Session Expired");
//			httpd_resp_send(req,(const char*)tempb, strlen(tempb));
//			return ESP_OK;
//		}
		xTimerReset(webTimer,0); //Start it again

#if SHAC>8
    long long shaint=1,gotsha=0;
#else
    uint32_t shaint=0,gotsha=1;
#endif

    buf_len = httpd_req_get_url_query_len(req) + 1;
    if (buf_len > 1)
    {
		sprintf(tempb,(char*)final_start,"nak","Invalid parameters");
        buf = (char*)malloc(buf_len);
        if(buf)
        {
			if (httpd_req_get_url_query_str(req, buf, buf_len) == ESP_OK)
			{
				if (httpd_query_key_value(buf, "challenge", param, sizeof(param)) == ESP_OK)
				{
					int ll=strlen(param);

					if(ll<SHAC)
					{
						sprintf(tempb,(char*)final_start,"nak","Invalid Challenge count");
						httpd_resp_send(req, tempb, strlen(tempb));
						return ESP_OK;
					}
					if(ll>SHAC)
						param[SHAC]=0; //cut it at 8 chars

#if SHAC>8
					gotsha=strtoull(param, NULL, 16);
#else
					gotsha=strtoul(param, NULL, 16);
#endif

					uint8_t *p=(uint8_t*)&shaint;				//little endian
					uint8_t * sh=(uint8_t*)&sharesult[SHAL];	//big endian

					for (int a=SHAL;a>=0;a--)	//big endian to little endian
						*(p++) = *(sh--);

				//	pprintf("Shatint %llu gotsha %llu\n",shaint,gotsha);

				//	if (shaint==gotsha)		//correct challenge
						if(1)
					{
						int van=0;

						for (int a=0;a<MAXDEVS;a++)
						{
							if(!theConf.configured[a])
							{	//not active erase configuration
								theConf.lock=false;		//allow further setup and launch task when rebooting
								theConf.beatsPerKw[a]=theConf.bornKwh[a]=0;
								bzero(theConf.medidor_id[a],sizeof(theConf.medidor_id[0]));
							}
							else
							{
								van++;
								strcpy(theConf.medidor_id[a],setupHost[a].meterid);
								time((time_t*)&theConf.bornDate[a]);
								theConf.beatsPerKw[a]			=setupHost[a].bpkwh;
								theConf.bornKwh[a]				=setupHost[a].startKwh;
							}
							if(van==MAXDEVS)
								theConf.lock=true;
							else
								theConf.lock=false;

							if(framFlag)
							{
								if(xSemaphoreTake(framSem, portMAX_DELAY/  portTICK_RATE_MS))
								{
									fram.formatMeter(a);
									fram.write_lifekwh(a,setupHost[a].startKwh);	// write to Fram. Its beginning of life KWH
									xSemaphoreGive(framSem);
								}
							}
				//			pprintf("Writing life Meter %d= %d\n",a,setupHost[a].startKwh);
						}
						bzero(tempb,1000);
						strcpy(theConf.mgrName,connserver);
						strcpy(theConf.mgrPass,connserver);
						// request Slot
						if(reserveSlot()==ESP_OK)
						{
							theConf.active=1;
							theConf.crypt=1;
							rebf=true;
							write_to_flash();
							sprintf(tempb,(char*)final_start,"ok","Success...rebooting in 3 secs, Login again");
						}
						else
						{
							theConf.active=0;
							sprintf(tempb,(char*)final_start,"nak","ConnMgr Authorization Denied");
						}
					}
					else
						sprintf(tempb,(char*)final_start,"nak","Invalid challenge");
				}
				else
					sprintf(tempb,(char*)final_start,"nak","Missing challenge");
			}
			free(buf);
        }
     }
    webState=wNONE; 	// force Login again
	httpd_resp_send(req, tempb, strlen(tempb));
	if(rebf)
	{
		xTaskCreate(&resetlater,"displ",2048,(void*)1, 4, NULL);	//wait and restart. Need to answer httpd before
	}
    return ESP_OK;
}

static bool getParam(const char *buf,const char *cualp,char *donde)
{
	if( httpd_query_key_value(buf, cualp, donde, 30)==ESP_OK)
		return true;
	else
		return false;
}


static esp_err_t conn_base(httpd_req_t *req)		//default page
{
//   if(!(webState==wNONE || webState==wMENU))
//    {
//		sprintf(tempb,(char*)final_start,"nak","Invalid Control Sequence or Session Expired");
//		httpd_resp_send(req,(const char*)tempb, strlen(tempb));
//		return ESP_OK;
//    }
	httpd_resp_set_hdr(req,"Cache-Control","private, max-age=86400");	//one day cache
	sendHtml(req,(char*)login_start,gwStr,theConf.meterName);
	webState=wLOGIN;
	xTimerStart(webTimer,0); //Start it

	return ESP_OK;
}

static esp_err_t conn_menu(httpd_req_t *req)
{
//   if(!(webState==wLOGIN ))
//	{
//		sprintf(tempb,(char*)final_start,"nak","Invalid Control Sequence or Session Expired");
//		httpd_resp_send(req,(const char*)tempb, strlen(tempb));
//		return ESP_OK;
//	}
	xTimerReset(webTimer,0); //Start it again

	httpd_resp_set_hdr(req,"Cache-Control","private, max-age=86400");	//one day cache
	sendHtml(req,(char*)menu_start,theConf.meterName,theConf.lock?"false":"true");
	webState=wMENU;
	return ESP_OK;
}

static esp_err_t conn_login(httpd_req_t *req)
{
	int buf_len;
		char *buf=NULL;
		char param[30];

//	   if(webState!=wLOGIN)
//		{
//			sprintf(tempb,(char*)final_start,"nak","Invalid Control Sequence or Session Expired");
//			httpd_resp_send(req,(const char*)tempb, strlen(tempb));
//			return ESP_OK;
//		}
		xTimerReset(webTimer,0); //Start it again

		buf_len = httpd_req_get_url_query_len(req) + 1;
		if (buf_len > 1)
		{
			buf = (char*)malloc(buf_len);
			if (httpd_req_get_url_query_str(req, buf, buf_len) == ESP_OK)
			{
				bzero(param,sizeof(param));
				if(getParam(buf,(char*)"user",param))
				{
					if(strlen(param)>=8)
					{
						strcpy(username,param);
						bzero(param,sizeof(param));
						if(getParam(buf,(char*)"password",param))
						{
							if(strlen(param)>=8)
							{
								strcpy(password,param);
								bzero(tempb,2000);
							}
						}
					}
				}
			}
		}
		sendHtml(req,(char*)menu_start,theConf.meterName,theConf.lock?"false":"true");
	webState=wMENU;
	return ESP_OK;
}


static esp_err_t conn_setup(httpd_req_t *req) //receives user/password and sends Conf Data
{
//	   if(webState!=wMENU)
//		{
//			sprintf(tempb,(char*)final_start,"nak","Invalid Control Sequence or Session Expired");
//			httpd_resp_send(req,(const char*)tempb, strlen(tempb));
//			return ESP_OK;
//		}
		xTimerReset(webTimer,0); //Start it again
		sendHtml(req,(const char*)setup_start,gwStr,theConf.meterName,
		theConf.meterName,theConf.sendDelay,
		theConf.medidor_id[0],theConf.beatsPerKw[0]==800?"Selected":"",theConf.beatsPerKw[0]==1000?"Selected":"",theConf.beatsPerKw[0]==1200?"Selected":"",theConf.beatsPerKw[0]==2000?"Selected":"",theConf.bornKwh[0],theConf.configured[0]?"checked":"",
		theConf.medidor_id[1],theConf.beatsPerKw[1]==800?"Selected":"",theConf.beatsPerKw[1]==1000?"Selected":"",theConf.beatsPerKw[1]==1200?"Selected":"",theConf.beatsPerKw[1]==2000?"Selected":"",theConf.bornKwh[1],theConf.configured[1]?"checked":"",
		theConf.medidor_id[2],theConf.beatsPerKw[2]==800?"Selected":"",theConf.beatsPerKw[2]==1000?"Selected":"",theConf.beatsPerKw[2]==1200?"Selected":"",theConf.beatsPerKw[2]==2000?"Selected":"",theConf.bornKwh[2],theConf.configured[2]?"checked":"",
		theConf.medidor_id[3],theConf.beatsPerKw[3]==800?"Selected":"",theConf.beatsPerKw[3]==1000?"Selected":"",theConf.beatsPerKw[3]==1200?"Selected":"",theConf.beatsPerKw[3]==2000?"Selected":"",theConf.bornKwh[3],theConf.configured[3]?"checked":"",
		theConf.medidor_id[4],theConf.beatsPerKw[4]==800?"Selected":"",theConf.beatsPerKw[4]==1000?"Selected":"",theConf.beatsPerKw[4]==1200?"Selected":"",theConf.beatsPerKw[4]==2000?"Selected":"",theConf.bornKwh[4],theConf.configured[4]?"checked":"",
				gwStr);
		webState=wSETUP;
	return ESP_OK;
}

static esp_err_t conn_status(httpd_req_t *req) //receives user/password and sends Conf Data
{
	char statesText[4][10];
	eTaskState states[4];

//	   if(webState!=wMENU)
//		{
//			sprintf(tempb,(char*)final_start,"nak","Invalid Control Sequence or Session Expired");
//			httpd_resp_send(req,(const char*)tempb, strlen(tempb));
//			return ESP_OK;
//		}

	xTimerReset(webTimer,0); //Start it again

	if(connHandle)
		states[0]=eTaskGetState(connHandle);
	else
		states[0]=(eTaskState)6;
	if(pcntHandle)
		states[1]=eTaskGetState(pcntHandle);
	else
		states[1]=(eTaskState)6;
	if(framHandle)
		states[2]=eTaskGetState(framHandle);
	else
		states[2]=(eTaskState)6;
	if(timeHandle)
		states[3]=eTaskGetState(timeHandle);
	else
		states[3]=(eTaskState)6;

	for (int a=0;a<4;a++)
	{
		switch(states[a])
		{
		case eRunning:
			strcpy(statesText[a],"RUN");
			break;
		case eReady:
			strcpy(statesText[a],"READY");
			break;
		case eBlocked:
			strcpy(statesText[a],"BLK");
			break;
		case eSuspended:
			strcpy(statesText[a],"SUSP");
			break;
		case eDeleted:
			strcpy(statesText[a],"DEL");
			break;
		default:
			strcpy(statesText[a],"DOWN");
			break;
		}
	}
	sendHtml(req,(const char*)status_start,theConf.meterName,
	theConf.meterName,theConf.sendDelay,theConf.active?"RUN":"SETUP",theConf.lock?"Y":"N",
	theConf.bootcount,theConf.lastResetCode,framFlag?"OK":"DOWN",conn?"ON":"OFF",
	statesText[0],statesText[1],statesText[2],statesText[3],
	strlen(theConf.medidor_id[0])?theConf.medidor_id[0]:"FREE",theConf.beatsPerKw[0],theMeters[0].curLife,theMeters[0].currentBeat,theConf.configured[0]?"ON":"OFF",
	strlen(theConf.medidor_id[1])?theConf.medidor_id[1]:"FREE",theConf.beatsPerKw[1],theMeters[1].curLife,theMeters[1].currentBeat,theConf.configured[1]?"ON":"OFF",
	strlen(theConf.medidor_id[2])?theConf.medidor_id[2]:"FREE",theConf.beatsPerKw[2],theMeters[2].curLife,theMeters[2].currentBeat,theConf.configured[2]?"ON":"OFF",
	strlen(theConf.medidor_id[3])?theConf.medidor_id[3]:"FREE",theConf.beatsPerKw[3],theMeters[3].curLife,theMeters[3].currentBeat,theConf.configured[3]?"ON":"OFF",
	strlen(theConf.medidor_id[4])?theConf.medidor_id[4]:"FREE",theConf.beatsPerKw[4],theMeters[4].curLife,theMeters[4].currentBeat,theConf.configured[4]?"ON":"OFF",
	gwStr);
	webState=wLOGIN;
	return ESP_OK;
}

static esp_err_t conn_stats(httpd_req_t *req) //receives user/password and sends Conf Data
{
	char txb[60],rxb[60],boot[60];
	struct tm * timeinfo;

	timeinfo = localtime (&stats.lastTx);
	strftime (txb,60," %D %T\n",timeinfo);
	timeinfo = localtime (&stats.lastRx);
	strftime (rxb,60," %D %T\n",timeinfo);
	timeinfo = localtime (&stats.bootTime);
	strftime (boot,60," %D %T\n",timeinfo);
	sendHtml(req,(const char*)stats_start,theConf.meterName,
	stats.txMsgSent,stats.txBytesSent,stats.txErrors,txb,
	stats.rxMsgIn,stats.rxBytesIn,stats.rxErrors,rxb,
	stats.conns,stats.disco,boot,esp_get_free_heap_size(),gwStr,gwStr);
	webState=wLOGIN;
	xTimerReset(webTimer,0); //Start it again

	return ESP_OK;
}

static esp_err_t conn_config_handler(httpd_req_t *req)
{
	    int buf_len;
		char param[32],tt[10];
		bool success=false;
	    char challengeSHA[50],aca[20];
	    char *buf=NULL;
	    if(webState==wSETUP)
	    {
	    	xTimerReset(webTimer,0); //Start it again

		buf_len = httpd_req_get_url_query_len(req) + 1;
		if (buf_len > 1)
		{
			buf = (char*)malloc(buf_len);
			if (httpd_req_get_url_query_str(req, buf, buf_len) == ESP_OK)
			{
				bzero(param,sizeof(param));
				if(getParam(buf,(char*)"conn",param))
				{
					sprintf(tempb,(char*)final_start,"nak","Invalid Paramters");
				   	strcpy(theConf.meterName,param);
					for (int a=0;a<MAXDEVS;a++)
					{
						sprintf(tt,"m%d",a+1);
						bzero(param,sizeof(param));
						if(getParam(buf,tt,param))
							if(strlen(param)>0)
								strcpy(setupHost[a].meterid,param);
							else
								continue;
						else
							continue;

						sprintf(tt,"b%d",a+1);
						bzero(param,sizeof(param));
						if(getParam(buf,tt,param))
							if(strlen(param)>0)
								setupHost[a].bpkwh=atoi(param);
							else
								continue;
						else
							continue;

						sprintf(tt,"k%d",a+1);
						bzero(param,sizeof(param));
						if(getParam(buf,tt,param))
							if(strlen(param)>0)
								setupHost[a].startKwh=atoi(param);
							else
								continue;
						else
							continue;

						sprintf(tt,"c%d",a+1);
						bzero(param,sizeof(param));
						if(getParam(buf,tt,param))
								theConf.configured[a]=1;
							else
								theConf.configured[a]=0;

						success=true;
					}
				}
			}
		}
	    }
	    else
	    {
			sprintf(tempb,(char*)final_start,"nak","Invalid Control Sequence or Session Expired");
			success=false;
	    }
		if(success)
		{
			if(getParam(buf,"crypt",param))
				global_crypt=true;
			else
				global_crypt=false;

		 	 sprintf(challengeSHA,"%s%u%06x",theConf.meterName,millis(),theMacNum);
			 shaMake(challengeSHA,strlen(challengeSHA),(uint8_t*)sharesult);
			 memcpy(aca,sharesult,16);
			 aca[16]=0;
			 shaMake(aca,16,(uint8_t*)sharesult);		//sharesult has the SHA to compare against when challenge received
			 size_t writ=0;
			 char newdest[100];
			 bzero(newdest,sizeof(newdest));

			 mbedtls_base64_encode((unsigned char*)newdest,sizeof(newdest),&writ,(unsigned char*)aca,16);
			 sendHtml(req,(char*)answer_start,theConf.meterName,gwStr,newdest);
			 webState=wCHALL;
			 return ESP_OK;
		}
		httpd_resp_send(req, tempb, strlen(tempb));

	    return ESP_OK;
}

void doScanBackground(void *pArg)
{
	wifi_config_t wifi_config;

	if(theConf.active)
	{
		strcpy(connserver,theConf.mgrName);
	//	printf("Copy connserver %s\n",connserver);
		vTaskDelete(NULL);			//already connected
	}
	bzero(connserver,sizeof(connserver));				//global name of ConnMgr
    if(scan(connserver)==ESP_OK)
    {
    	memset(&wifi_config,0,sizeof(wifi_config));//very important
    	strcpy((char*)wifi_config.sta.ssid,connserver);
		strcpy((char*)wifi_config.sta.password,connserver);
		ESP_ERROR_CHECK(esp_wifi_set_config(ESP_IF_WIFI_STA, &wifi_config));
		ESP_ERROR_CHECK(esp_wifi_connect());
    }
	vTaskDelete(NULL);
}

void start_webserver(void *pArg)
{
    httpd_ssl_config_t config = HTTPD_SSL_CONFIG_DEFAULT();
    httpd_handle_t 		server;

    config.httpd.max_uri_handlers=MAXURLS;
    config.httpd.max_open_sockets=1;		//just one connection
    config.httpd.stack_size=20000;

    extern const unsigned char cacert_pem_start[] asm("_binary_cacert_pem_start");
    extern const unsigned char cacert_pem_end[]   asm("_binary_cacert_pem_end");
    config.cacert_pem = cacert_pem_start;
    config.cacert_len = cacert_pem_end - cacert_pem_start;
    extern const unsigned char pprvtkey_pem_start[] asm("_binary_wprvtkey_pem_start");
    extern const unsigned char pprvtkey_pem_end[]   asm("_binary_wprvtkey_pem_end");
    config.prvtkey_pem = pprvtkey_pem_start;
    config.prvtkey_len = pprvtkey_pem_end - pprvtkey_pem_start;

	esp_netif_ip_info_t ipInfo1;
    esp_netif_get_ip_info(wifiAP, &ipInfo1);
    sprintf(gwStr,"https://" IPSTR,IP2STR(&ipInfo1.gw));		//so that href are with whatever GW its connected to


    if(theConf.traceflag & (1<<WEBD))
    	printf("%sStarting server on port:%d\n",WEBDT, config.httpd.server_port);

    int ret=httpd_ssl_start(&server, &config);	//global wserver to stop it if Firmware called

    if (ret == ESP_OK) {
    	webState=wNONE;

    	xTaskCreate(&doScanBackground,"web",4096,(void*)1, 4, NULL);
    	webTimer=xTimerCreate("Monitor",600000 /portTICK_PERIOD_MS,pdTRUE,NULL,&monitorCallback);	//10 minutes

    	bzero(&setupHost,sizeof(setupHost));
        // Set URI handlers

 	   init_urls();

        // Set URI handlers
       	for(int a=0;a<MAXURLS;a++)
       	{
       		urls[a].method=HTTP_GET;
       		urls[a].user_ctx=NULL;
       		httpd_register_uri_handler(server, &urls[a]);;
       	}
    }

#ifdef DEBUGX
    if(theConf.traceflag & (1<<WEBD))
    	pprintf("WebServer Started\n");
#endif
    vTaskDelete(NULL);		//we are done
}
